#!perl

####
# RegExpServices.pl:  Writes the results of a regular expression operation
# to STDOUT.
#
#####
#
# Usage:
#
#     perl RegExpServices.pl {split|exec|test|sub} <inputFilePath>
#
# The script processes the command-line arguments, and if they are valid,
# the script then reads additional data from the specified input file.
# For further details, see the readme file (Readme.txt).
#
####
#
# Exit codes:
#
#     0 - No problem!
#     1 - Invalid list of command-line arguments.
#     2 - Invalid file path, or couldn't read the file.
#     3 - Invalid file contents.
#     4 - Invalid regular expression.
#
#   NOTE:  If the regular expression is not valid, the script writes the
#   error message generated by Perl to STDOUT.
####
use strict;
use English "-no_match_vars";
use Scalar::Util ("looks_like_number");

# Constants:
my $POUND = "#";
my $SQ    = "'";
my $DQ    = "\"";

my $INCLUDE_TRAILING_EMPTY_FIELDS = -1;

# Global variables:
my @args = @ARGV;

my $operation;
my $input_path;
my $input;
my $pattern;
my $compiled_reg;
my $target_string;
my $replacement_string;
my $result_string;


# Main script:

# Capture the command-line arguments:
unless (scalar @args) {
    exit(1);
}

unless (scalar(@args) > 1) {
    exit(1);
}

if ($args[0] =~ /\A(?:split|exec|test|sub)\z/is) {
    $operation = lc($args[0]);
}
else {
    exit(1);
}

$input_path = $args[1];

# Extract $pattern, $target_string, and (if needed) $replacement_string
# from the input file:
if (open my $INPUT_FILE, $input_path) {
    binmode $INPUT_FILE;
    $input = join "", <$INPUT_FILE>;
    close $INPUT_FILE;
}
else {
    exit(2);
}

process_input();

# Perform the regular expression operation:
eval { $compiled_reg = qr/$pattern/s };

if ($EVAL_ERROR) {
    print $EVAL_ERROR;
    exit(4);
}

if ($operation eq "split") {
    do_split();
}
elsif ($operation eq "test") {
    do_test();
}
elsif ($operation eq "sub") {
    do_substitution();
}
else {
    get_matches();
}

# Write the results to STDOUT:
binmode STDOUT;
print $result_string;

# Exit:
exit(0);

# End main script


# Subroutines:

# The process_input() subroutine extracts the pattern, the target string,
# and (if needed) the replacement string from $input.
sub process_input {
    if ($input =~ /(?:\A|(?<=\n))Pattern=([^\r\n]+)/s) {
        ($pattern) = $input =~ /(?:\A|(?<=\n))Pattern=([^\r\n]+)/gs;
    }
    else {
        exit(3);
    }

    if ($input =~ /(?:\A|(?<=\n))TargetString=([^\r\n]+)/s) {
        ($target_string)
          = $input =~ /(?:\A|(?<=\n))TargetString=([^\r\n]+)/gs;

        $target_string =~ s/((?<!\\)(?:\\\\)*)\\r/$1\r/gs;
        $target_string =~ s/((?<!\\)(?:\\\\)*)\\n/$1\n/gs;
        $target_string =~ s/\\\\/\\/gs;
    }
    else {
        exit(3);
    }

    if ($operation eq "sub") {
        if ($input =~ /(?:\A|(?<=\n))ReplacementString=([^\r\n]*)/s) {
            ($replacement_string)
              = $input =~ /(?:\A|(?<=\n))ReplacementString=([^\r\n]*)/gs;

            $replacement_string =~ s/((?<!\\)(?:\\\\)*)\\r/$1\r/gs;
            $replacement_string =~ s/((?<!\\)(?:\\\\)*)\\n/$1\n/gs;
            $replacement_string =~ s/\\\\/\\/gs;
        }
        else {
            exit(3);
        }
    }
    else {
        $replacement_string = "";
    }
}

# The do_split() subroutine puts the results of a split() operation into
# $result_string.  Trailing empty fields are included in the results.

# In order to allow another program to read just a single line of input
# coming from this script's standard output stream, this subroutine uses \r
# to represent a carriage return, \n to represent a line feed, and \\ to
# represent a literal backslash.  The subroutine uses \z to terminate each
# array element.
sub do_split {
    my @pieces = split($compiled_reg, $target_string,
      $INCLUDE_TRAILING_EMPTY_FIELDS);

    $result_string = "";

    for (@pieces) {
        s/\\/\\\\/gs;
        s/\r/\\r/gs;
        s/\n/\\n/gs;
        $result_string .= "$_\\z";
    }
}

# The do_test() subroutine puts the results of a regular-expression test
# into $result_string.
sub do_test {
    if ($target_string =~ $compiled_reg) {
        $result_string = "1";
    }
    else {
        $result_string = "0";
    }
}

# The do_substitution() subroutine puts the results of a regular-expression
# substitution into $result_string.  The subroutine uses \r to represent a
# carriage return, \n to represent a line feed, and \\ to represent a
# literal backslash.
#
#   NOTE:  This subroutine treats dollar signs within the replacement string
#   literally.  Variables such as $1, $2, and so forth are not supported.
sub do_substitution {
    $result_string = $target_string;
    $result_string =~ s/$compiled_reg/$replacement_string/g;
    $result_string =~ s/\\/\\\\/gs;
    $result_string =~ s/\r/\\r/gs;
    $result_string =~ s/\n/\\n/gs;
}

# The get_matches() subroutine captures in an array all pattern matches
# that occur in the target string.  The resulting array does not include
# extra elements to hold the results of capturing parentheses.
#
# The subroutine places the array in $result_string, using \r to
# represent a carriage return, \n to represent a line feed, and \\ to
# represent a literal backslash.  The subroutine uses \z to terminate
# each array element.
#
# If there are no matches, the subroutine writes a zero (0) to STDOUT
# with no terminating \z or end-of-line characters.
sub get_matches {
    my @matches;

    while ($target_string =~ /$compiled_reg/g) {
        push @matches, substr($target_string, $-[0], $+[0] - $-[0]);
    }

    unless (scalar @matches) {
        $result_string = "0";
        return;
    }

    $result_string = "";

    for (@matches) {
        s/\\/\\\\/gs;
        s/\r/\\r/gs;
        s/\n/\\n/gs;
        $result_string .= "$_\\z";
    }
}
